/*
	vpdpart.c

	original code: SASANO Takayoshi
	--- public domain, no warranty.
*/

#include <basic.h>
#include <bstdio.h>
#include "vpdpart.h"

#ifdef	DEBUG
#define	D(x)	printf x
#else
#define	D(x)	/* */
#endif

#define	HEADER_POS(sectsize)	((sectsize) - sizeof(VPD_HEADER))

/* $B%X%C%@:n@.(B */
EXPORT	W	create_header(VP buf, W maxentry, W sectsize)
{
	W	err, pos, offset;
	VPD_HEADER	*h;

	/* $B%X%C%@$H6h2h%F!<%V%k$N0LCV7W;;(B */
	pos    = HEADER_POS(sectsize);
	offset = pos - sizeof(VPD_PARTINFO) * maxentry;

	if (offset < 0) {
		err = PartInfoError(BadEntry);
		goto fin0;
	}

	/* $B%;%/%?$NFbMF$r>C5n(B */
	memset(buf, 0, sectsize);

	/* $B%X%C%@>pJs$N:n@.(B */
	h = (VPD_HEADER *)(buf + pos);
	h->signature = LittleEndW(VPD_SIGNATURE0);
	h->blocksize = LittleEndW(sectsize);
	h->offset    = LittleEndW(offset);
	h->entry     = LittleEndW(maxentry);

	err = ER_OK;
 fin0:
	return err;
}

/* $B%X%C%@3NG'(B */
EXPORT	VPD_HEADER	*check_header(VP buf, W maxentry, W sectsize)
{
	W	pos, signature, blocksize, offset, entry;
	VPD_HEADER	*h;

	/* $B%X%C%@$N0LCV7W;;(B */
	pos = HEADER_POS(sectsize);
	h   = (VPD_HEADER *)(buf + pos);

	/* $B%X%C%@$N<h$j9~$_(B */
	signature = LittleEndW(h->signature);
	blocksize = LittleEndW(h->blocksize);
	offset    = LittleEndW(h->offset);
	entry     = LittleEndW(h->entry);

	/* $B%7%0%M%A%c$N3NG'(B */
	if (signature != VPD_SIGNATURE0) {
		D(("load_partinfo: signature %#x\n", signature));
		goto fin0;
	}

	/* $B%;%/%?%5%$%:$N3NG'(B */
	if (blocksize != sectsize) {
		D(("load_partinfo: blocksize %d\n", blocksize));
		goto fin0;
	}

	/* $B6h2h%F!<%V%k$N0LCV3NG'(B */
	if (offset < 0 || offset >= pos) {
		D(("load_partinfo: offset %#x\n", offset));
		goto fin0;
	}

	/* $B6h2h%(%s%H%j?t$N3NG'(B */
	if (entry < 1 || entry > maxentry ||
	    entry > (pos - offset) / sizeof(VPD_PARTINFO)) {
		D(("load_partinfo: entry %d\n", entry));
		goto fin0;
	}

	/* $B%(%i!<$J$7(B */
	return h;

 fin0:
	/* $B%(%i!<$"$j(B */
	return NULL;
}

/* $B6h2h>pJs$N=q$-=P$7(B */
EXPORT	W	export_partinfo(VP buf, VPD_PARTINFO *info, W maxentry, W sectsize)
{
	W	err, i, entry;
	VPD_HEADER	*h;
	VPD_PARTINFO	*p;

	/* $B%X%C%@$N3NG'(B */
	h = check_header(buf, maxentry, sectsize);
	if (h == NULL) {
		err = PartInfoError(BadHeader);
		goto fin0;
	}

	/* $B6h2h%F!<%V%k$N0LCV7W;;(B */
	p = (VPD_PARTINFO *)(buf + LittleEndW(h->offset));

	/* $B6h2h>pJs$N=q$-=P$7(B */
	entry = LittleEndW(h->entry);
	for (i = 0; i < entry; i++, p++, info++) {
		p->systemid   = LittleEndW(info->systemid);
		p->startblock = LittleEndW(info->startblock);
		p->endblock   = LittleEndW(info->endblock);
		p->reserved   = LittleEndW(info->reserved);
	}

	err = entry;
 fin0:
	return err;
}

/* $B6h2h>pJs$N<h$j9~$_(B */
EXPORT	W	import_partinfo(VP buf, VPD_PARTINFO *info, W maxentry, W sectsize)
{
	W	err, i, entry;
	VPD_HEADER	*h;
	VPD_PARTINFO	*p;

	/* $B%X%C%@$N3NG'(B */
	h = check_header(buf, maxentry, sectsize);
	if (h == NULL) {
		err = PartInfoError(BadHeader);
		goto fin0;
	}

	/* $B<h$j9~$_@h$N%/%j%"(B */
	memset(info, 0, sizeof(VPD_PARTINFO) * maxentry);

	/* $B6h2h%F!<%V%k$N0LCV7W;;(B */
	p = (VPD_PARTINFO *)(buf + LittleEndW(h->offset));

	/* $B6h2h>pJs$N<h$j9~$_(B */
	entry = LittleEndW(h->entry);
	for (i = 0; i < entry; i++, info++, p++) {
		info->systemid   = LittleEndW(p->systemid);
		info->startblock = LittleEndW(p->startblock);
		info->endblock   = LittleEndW(p->endblock);
		info->reserved   = LittleEndW(p->reserved);
	}

	err = entry;
 fin0:
	return err;
}

/* $BIT@5$J6h2h>pJs$NL58z2=(B */
EXPORT	W	check_partinfo(VPD_PARTINFO *info, W maxentry, W start, W end)
{
	W	err, i, j;

	/* check 1: $B4{$K%(%i!<07$$$K$J$C$F$$$k6h2h$OL$;HMQ(B(0)$B$H$9$k(B */
	for (i = 0; i < maxentry; i++) {
		if (info[i].systemid < 0) info[i].systemid = 0;
	}

	/* check 2: $B4pK\E*$J%Q%i%a!<%?$N%A%'%C%/(B */
	for (i = 0; i < maxentry; i++) {
		if (info[i].systemid <= 0) continue;

		/* $B$I$N%Q%i%a!<%?$,$*$+$7$$$+%A%'%C%/(B */
		if (info[i].startblock < start ||
		    info[i].startblock > end) {
			info[i].systemid = PartInfoError(BadStartBlock);

		} else if (info[i].endblock < start ||
			   info[i].endblock > end) {
			info[i].systemid = PartInfoError(BadEndBlock);

		} else if (info[i].startblock > info[i].endblock) {
			info[i].systemid = PartInfoError(BadStartEndBlock);

		}
	}

	/* check 3: $B6h2h>pJs$N>WFM%A%'%C%/(B */
	for (i = 0; i < maxentry; i++) {
		if (info[i].systemid <= 0) continue;

		/* $B$h$j<c$$HV9f$N%(%s%H%j$H>WFM$7$F$$$J$$$+%A%'%C%/(B */
		for (j = 0; j < i; j++) {
			if (info[j].systemid <= 0) continue;

			if ((info[i].startblock >= info[j].startblock &&
			     info[i].startblock <= info[j].endblock) ||
			    (info[i].endblock   >= info[j].startblock &&
			     info[i].endblock   <= info[j].endblock)) {
				break;	/* $B>WFM$7$?(B */
			}
		}

		/* $B>WFM$7$?>l9g!">WFM$7$?B&$*$h$S;D$jA4$F$N%(%s%H%j$rL58z2=(B */
		if (j != i) {
			for (j = i; j < maxentry; j++) {
				if (info[j].systemid > 0) {
					info[j].systemid =
						PartInfoError(BadEntry);
				}
			}
			break;
		}
	}

	/* check 4: $B%(%i!<$K$h$C$FL58z2=$5$l$?%(%s%H%j$NM-L5$r%A%'%C%/(B */
	err = ER_OK;
	for (i = 0; i < maxentry; i++) {
		if (info[i].systemid < 0) {
			err = info[i].systemid;
			break;
		}
	}

	return err;
}

#ifdef	DEBUG
/* $B6h2h>pJs$NI=<(!J%G%P%C%0MQ!K(B */
EXPORT	void	disp_partinfo(VPD_PARTINFO *info, W entry)
{
	W	i;

	for (i = 0; i < entry; i++, info++) {
		printf("[%2d] (%08x) %08x %08x %08x\n", i,
		       info->systemid, info->startblock,
		       info->endblock, info->reserved);
	}

	return;
}
#endif
